"""Concatenate MRI time series data.

Usage:
    tsconcat bids_dir output_dir group [options]

Arguments:
    bids_dir (str): Input BIDS folder path.
    output_dir (str): Output BIDS folder path.
    analysis_level (str): Processing stage, must be "group".
"""



"""Concatenate 4D Nifti files."""

import pathlib as pl
from os import PathLike
from typing import Iterable, List, Union

import nibabel as nib
import numpy as np


def concat_nifti1_4d(paths: Iterable[Union[str, PathLike]], out_path: Union[str, PathLike]) -> None:
    """Concat 4D Nifti files.

    Args:
        paths: Paths to 4D Nifti files.
        out_path: Path to output 4D Nifti file.

    Raises:
        Exception: If affines are not equal.
        Exception: If path list is empty.
    """
    paths = [pl.Path(p) for p in paths]

    if len(paths) == 0:
        raise Exception("Empty path list.")

    img_handles: List[nib.nifti1.Nifti1Image] = [nib.nifti1.Nifti1Image.load(p) for p in paths]  # type: ignore

    img_shapes: List[tuple] = [img.shape for img in img_handles]

    # All 4D
    if not np.all([len(shape) == 4 for shape in img_shapes]):
        raise Exception("Images must be 4D.")

    # First 3 dimensions must be equal
    if not np.all([img_shapes[0][:-1] == shape[:-1] for shape in img_shapes]):
        raise Exception("Spatial shapes must be equal.")

    img_affines: List[np.ndarray] = [img.affine for img in img_handles]

    affs = np.asarray(img_affines)
    if not np.all(affs == affs[0, :, :]):
        raise Exception("Affines must be equal.")

    # Should this use the _dataobj to preserve integer types?
    img_arrays = [img.get_fdata() for img in img_handles]

    concat = np.concatenate(img_arrays, axis=3)

    concat_n1 = nib.nifti1.Nifti1Image(concat, img_affines[0])
    nib.nifti1.save(concat_n1, out_path)



"""Utility functions for tsconcat."""

import argparse
import datetime
import pathlib as pl
import time
from contextlib import contextmanager
from typing import Generator, List

import bids2table.table
import pandas as pd

from tsconcat.b2t_columns import B2tColumn


@contextmanager
def timeprint(title: str) -> Generator[None, None, None]:
    """Context manager to print the time elapsed between entering and exiting the context.

    Args:
        title: Title to be printed before and after the context.
    """
    print(f"Start: {title}")
    start = time.perf_counter()
    yield
    duration = time.perf_counter() - start
    print(f"Done: {title} - {datetime.timedelta(seconds=duration)}")


def build_bidsapp_group_parser(*args, **kwargs) -> argparse.ArgumentParser:  # noqa
    """Build a parser skeleton for the BIDS App group level.

    Args:
        args: Positional arguments to be passed to ArgumentParser costructor.
        kwargs: Keyword arguments to be passed to ArgumentParser costructor.
    """
    parser = argparse.ArgumentParser(*args, **kwargs)
    parser.add_argument(
        "bids_dir",
        action="store",
        type=pl.Path,
        help="Input BIDS folder path.",
    )
    parser.add_argument(
        "output_dir",
        action="store",
        type=pl.Path,
        help="Output BIDS folder path.",
    )
    parser.add_argument("analysis_level", default="group", choices=["group"], help='Processing stage, must be "group".')
    return parser


def file_paths_from_b2table(df: pd.DataFrame, include_sidecars: bool = False, inplace: bool = False) -> List[pl.Path]:
    """Generate list of filepaths from bids2table dataframe."""
    # b2t crashes if sidecar is not None
    if not inplace:
        df = df.copy()
    df[B2tColumn.MetaJson] = None

    paths = list(df.apply(func=lambda row: bids2table.table.join_bids_path(row), axis=1).values)

    if include_sidecars:
        sidecar_paths = list(
            df.apply(
                func=lambda row: bids2table.table.join_bids_path({**row, B2tColumn.FileExtension: ".json"}),
                axis=1,
            ).values
        )
        return paths + sidecar_paths

    return paths


def file_path_from_b2table_row(row: pd.Series, inplace: bool = False, sidecar: bool = False) -> pl.Path:
    """Generate list of filepaths from bids2table dataframe."""
    # b2t crashes if sidecar is not None
    if not inplace:
        row = row.copy()
    row[B2tColumn.MetaJson] = None

    if sidecar:
        row = {**row, B2tColumn.FileExtension: ".json"}

    return bids2table.table.join_bids_path(row)


def sidecar_path_from_b2table_row(row: pd.Series, inplace: bool = False) -> pl.Path:
    """Generate list of filepaths from bids2table dataframe."""
    # b2t crashes if sidecar is not None
    if not inplace:
        row = row.copy()
    row[B2tColumn.MetaJson] = None

    return bids2table.table.join_bids_path({**row, B2tColumn.FileExtension: ".json"})



"""Print a tree from a list of paths."""

import pathlib as pl
from typing import Dict, List, Optional, Sequence, Union

from rich.console import Console

console = Console(highlight=False)


def tree_from_paths(paths: Sequence[Union[str, pl.Path]]) -> dict:
    """Build a tree from a list of paths."""
    tree: Dict[str, dict] = {}
    for p in (pl.Path(p) for p in paths):
        parts = list(p.parts)

        node = tree
        while True:
            if len(parts) == 0:
                break
            segment = parts.pop(0)

            child = node.get(segment, None)

            if child is None:
                child = {}
                node[segment] = child

            node = child
    return tree


def tree_collapse(tree: dict) -> dict:
    """Collapse a tree by merging single-child nodes."""
    stack = [tree]
    while True:
        if len(stack) == 0:
            break
        node = stack.pop(0)

        updated_node = False

        for node_name, child in node.items():
            if len(child) == 1:
                child_name, childchild = tuple(child.items())[0]
                del node[node_name]
                node[str(pl.Path(node_name) / child_name)] = childchild
                updated_node = True
                break

        if updated_node:
            stack.insert(0, node)
        else:
            for child in node.values():
                stack.append(child)

    return tree


INDENT_T = "├── "
INDENT_L = "└── "
INDENT_O = "    "
INDENT_I = "│   "
RAINBOW_COLORS = ["red", "green", "blue", "magenta", "cyan"]


def tree_print(tree: dict, indents: Optional[List[bool]] = None) -> None:
    """Print a tree."""
    indents = [] if indents is None else indents

    for i, (k, v) in enumerate(tree.items()):
        is_last_child = i == len(tree) - 1

        indent_list = [
            ((INDENT_L if is_last_child else INDENT_T) if level == len(indents) else INDENT_I) if ind else INDENT_O
            for level, ind in enumerate(indents + [True])
            if level > 0
        ]

        indent_list = [
            f"[{RAINBOW_COLORS[level % len(RAINBOW_COLORS)]}]{ind}[/]" for level, ind in enumerate(indent_list)
        ]

        total_indent = "".join(indent_list)
        console.print(f"{total_indent}{k}")
        tree_print(v, indents + [not is_last_child])


def pretreeprint(paths: Sequence[Union[str, pl.Path]]) -> None:
    """Print a tree from a list of paths."""
    tree_print(tree_collapse(tree_from_paths(paths)))


if __name__ == "__main__":
    from glob import glob

    pretreeprint(list(glob("../../**", recursive=True)))



"""CLI for tsconcat."""

import argparse
import json
import pathlib as pl
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, List, Optional

import bids2table.table
import elbow.dtypes  # noqa  makes pandas load json types as dicts from parquet
import pandas as pd

from tsconcat.b2t_columns import B2tColumn

from .concat import concat_nifti1_4d
from .pretreeprint import pretreeprint
from .utils import (
    build_bidsapp_group_parser,
    file_path_from_b2table_row,
    file_paths_from_b2table,
    sidecar_path_from_b2table_row,
)

REDUCE_COLUMNS = [
    B2tColumn.Dataset,
    B2tColumn.Subject,
    B2tColumn.Session,
    B2tColumn.Run,
    B2tColumn.Description,
    B2tColumn.Space,
]
REDUCE_COLUMNS_SET = set(REDUCE_COLUMNS)
REDUCE_COLUMNS_ALIAS = {
    "dataset": B2tColumn.Dataset,
    "sub": B2tColumn.Subject,
    "subject": B2tColumn.Subject,
    "ses": B2tColumn.Session,
    "session": B2tColumn.Session,
    "run": B2tColumn.Run,
}


def _reduce_op(
    df: pd.DataFrame,
    group_by: List[str],
    inplace: bool = False,
    group_callback: Optional[Callable[[pd.DataFrame], None]] = None,
) -> pd.DataFrame:
    """Reduce dataframe to one row per group."""
    if not inplace:
        df = df.copy()

    group_by_set = set(group_by)

    unknown_cols = list(group_by_set - REDUCE_COLUMNS_SET)
    if len(unknown_cols) > 0:
        raise Exception(f"Unknown columns: {unknown_cols}")
    del unknown_cols

    df = df.sort_values(by=REDUCE_COLUMNS)

    group_by_columns = list(REDUCE_COLUMNS_SET - group_by_set)
    # Duplicate group_by columns in df to retain them
    # This used to be the default behavior of groupby.apply, but it was deprecated in pandas 2.2
    for col in group_by_columns:
        df[f"temp_copy_{col}"] = df[col]
    grouped = df.groupby(by=group_by_columns, dropna=False)

    def _func_reduce(df_group: pd.DataFrame) -> Optional[pd.Series]:
        if df_group.shape[0] == 0:
            print("empty group")
            return None

        # Name them back to normal
        df_group = df_group.rename(columns={f"temp_copy_{col}": col for col in group_by_columns})

        first_row: pd.Series = df_group.iloc[0]

        for group_label in group_by:
            first_row = first_row.copy()
            first_row[group_label] = None  # todo does vectorized work here?

        if group_callback is not None:
            group_callback(df_group)

        return first_row

    df_reduced = grouped.apply(func=_func_reduce, include_groups=False)
    df_reduced = df_reduced.reset_index(drop=False)

    return df_reduced


def _read_if_parquet(p: pl.Path, *args, **kwargs) -> Optional[pd.DataFrame]:  # noqa
    try:
        return pd.read_parquet(p, *args, **kwargs)
    except:  # noqa
        return None


def build_parser_tsconcat_bids_app() -> argparse.ArgumentParser:
    """Build parser for tsconcat BIDS app."""
    parser = build_bidsapp_group_parser(prog="ba-tsconcat", description="Concatenate MRI timeseries.")

    parser.add_argument(
        "-c",
        "--concat",
        type=str,
        help=f"Concat across. Can be any combination of {', '.join(REDUCE_COLUMNS_ALIAS.keys())} separated by spaces. "
        f"Output data will be grouped by the set difference.",
        default="ses",
    )

    parser.add_argument(
        "-d",
        "--dry_run",
        action="store_true",
        help="Dry run. Print output directory structure instead of actually doing something. "
        "If this is enabled 'bids_dir' may be a path to a bids2table parquet directory.",
        default=False,
    )

    parser.add_argument(
        "-f",
        "--fake",
        action="store_true",
        help="Fake output. Output a bids2table parquet directory instead of actually doing something.",
        default=False,
    )

    # workers
    parser.add_argument(
        "-w",
        "--workers",
        type=int,
        help="Number of workers for bids2table. Default is 1.",
        default=1,
    )

    return parser


@dataclass
class TSConcatSettings:
    """Settings for tsconcat."""

    input_dir: pl.Path
    output_dir: pl.Path
    concat_labels: List[str]
    dry_run: bool
    fake: bool
    workers: int

    @classmethod
    def from_args(cls, args: Any) -> "TSConcatSettings":  # noqa
        concat_labels = args.concat.split(" ")
        concat_set = set(REDUCE_COLUMNS_ALIAS.get(g, g) for g in concat_labels)

        return cls(
            input_dir=args.bids_dir,
            output_dir=args.output_dir,
            concat_labels=list(concat_set),
            dry_run=args.dry_run or args.fake,
            fake=args.fake,
            workers=args.workers,
        )


def main() -> None:
    """Concatenate MRI timeseries."""
    settings = TSConcatSettings.from_args(build_parser_tsconcat_bids_app().parse_args())

    if not settings.input_dir.exists():
        raise Exception("Input directory does not exist.")

    if settings.dry_run and (df := _read_if_parquet(settings.input_dir)) is not None:
        df = bids2table.table.flat_to_multi_columns(df)
    else:
        df = bids2table.bids2table(settings.input_dir, workers=settings.workers, with_meta=False)

        if df.shape[0] == 0:
            raise Exception("Empty BIDS dataset")

    if not settings.dry_run:
        settings.output_dir.mkdir(parents=True, exist_ok=True)

    df_bold = df.query(
        f"{B2tColumn.DataType} == 'func' and "
        f"{B2tColumn.FileExtension} == '.nii.gz' and "
        f"{B2tColumn.Suffix} == 'bold'"  # and "
        # "ent__desc == 'preproc' and "
        # "ent__space == 'MNI152NLin6ASym'"
    )  # fmt: skip

    if df_bold.shape[0] == 0:
        raise Exception("No BOLD files found")

    def _process_group(df_group: pd.DataFrame) -> None:
        group_identifiers = df_group.iloc[0][list(REDUCE_COLUMNS_SET - set(settings.concat_labels))].to_dict()
        print(f"Process group: {group_identifiers}")

        first_row: pd.Series = df_group.iloc[0]

        for group_label in settings.concat_labels:
            first_row = first_row.copy()
            first_row[group_label] = None

        # Generate file

        file_path = settings.output_dir / file_path_from_b2table_row(first_row)

        file_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            concat_nifti1_4d(paths=df_group.finfo__file_path.values, out_path=file_path)
        except Exception as e:
            print(f"Warning: Could not concat group, skipping:\n{e}")
            return

        # Generate sidecar

        sidecar_path = settings.output_dir / sidecar_path_from_b2table_row(first_row)
        sidecar_contents = first_row[B2tColumn.MetaJson]  # TODO: Maybe add list of files that were concatenated?
        with open(sidecar_path, "w", encoding="utf-8") as fp:
            json.dump(sidecar_contents, fp)

    df_reduced_bold = _reduce_op(
        df_bold,
        group_by=settings.concat_labels,
        group_callback=None if settings.dry_run else _process_group,
    )

    if settings.fake:
        # You need this one, believe me
        df_reduced_bold = df_reduced_bold.astype(
            {B2tColumn.MetaJson: "json", B2tColumn.DatasetDescription: "json", B2tColumn.ExtraEntities: "json"}
        )
        df_reduced_bold.to_parquet(settings.output_dir)

    filepaths = file_paths_from_b2table(df_reduced_bold, include_sidecars=True)
    pretreeprint(filepaths)


if __name__ == "__main__":
    main()



"""bids2table column name constants."""

from abc import ABC


class B2tColumn(ABC):
    """bids2table column name constants."""

    Dataset = "ds__dataset"
    DatasetDescription = "ds__dataset_description"

    DataType = "ent__datatype"
    """i.e. 'func', 'anat', 'dwi', 'fmap'"""
    Subject = "ent__sub"
    Session = "ent__ses"
    Run = "ent__run"
    Description = "ent__desc"
    Space = "ent__space"
    ExtraEntities = "ent__extra_entities"
    FileExtension = "ent__ext"
    Suffix = "ent__suffix"

    FilePath = "finfo__file_path"

    MetaJson = "meta__json"
    """JSON sidecar contents"""
